(************************************)
(*** Identify Disjunctive formula ***)
(************************************)

open Types;;
open Valuation;;

(* Separate implication expressions of each hybridConj of the super_valuation type 
A list is created with a tuple of the form 
(valuation , (hybridConj without "disjunctive formula" , hybridConj with "disjunctive formula"))
The hybrid condition HNeg and Hop(Impl, _, _) are transformed to disappear *)
let separate_disj sval cl =
  let neg_comp = function
    (* Negation of a comparator *)
    | Eq -> NEq
    | NEq -> Eq
    | LE -> GT
    | LT -> GE
    | GE -> LT
    | GT -> LE 
    in
  let neg_operator = function
    | And -> Or
    | Or -> And
    | Impl -> raise(Invalid_argument "neg_operator: should analyse And/Or operators only")
    in
  let rec reverse_comp h =
    match h with
      | HAtom(comp, h1, h2) -> HAtom((neg_comp comp), h1, h2)
      | HDAtom(comp, h1, h2) -> HDAtom((neg_comp comp), h1, h2)
      | HOp(op, HAtom(comp1,PiExit(v1,i1),float1), HAtom(comp2,PiExit(v2,i2),float2)) -> HOp((neg_operator op), HAtom(Eq, PiExit(v1,i1), HConstant(0.0)), HAtom(Eq, PiExit(v2,i2), HConstant(1.0)))
      (* The case HOp only includes the continuous disjunctions and so acts on the fractional parts. 
      We give a specific result of this case since π_v ≤ 0 means π_v = 0 *)
      | HOp(op, h1, h2) -> HOp((neg_operator op), (reverse_comp h1), (reverse_comp h2))
      | other -> other
    in
  let rec replace_neg h = (* Replace the neg constructor *)
    match h with
      | HNeg(neg, h1) -> reverse_comp h1
      (* | HOp(op, h1, h2) -> if (op = Impl) then (HOp(Or, HNeg(Neg,h1), h2)) else (raise(Invalid_argument "replace_neg: Error")) *)
      | HOp(Impl, h1, h2) -> HOp(Or, (reverse_comp h1), h2)
      | HOp(And, h1, h2) -> HOp(And, h1, h2)
      | HOp(Or, h1, h2) -> HOp(Or, h1, h2)
      | _ -> raise(Invalid_argument "replace_neg: error to select the disjunctions")
    in
  let rec find_disj h b = (* Find the Impl operator and the And operator included in neg constructor  *)
    match h with 
      | HOp(op, hc1, hc2) -> if ((op = Impl && b) || (op = And && b = false)) then (true) else (false)
        (* Above we only keep ¬(A∧B) and (A⇒B). ¬(A⇒B) can be replaced by (A∧¬B) *)
      | HNeg(neg, h1) -> find_disj h1 false 
      | otherConstructor -> false
    in
  let rec cel_in_list ht cl =
    match cl with
      | [] -> false
      | h :: t -> if (ht = h) then (true) else (cel_in_list ht t)
    in
  let rec hterm_in_list ht cl =
    match ht with
      | PiEntrance(_,_) | PiExit(_,_) | HConstant(_) | HTime(_) -> true
      | Celerity(v,w,n) -> cel_in_list (Celerity(v,w,n)) cl
      | HTermRelation(s, ht1, ht2) -> hterm_in_list ht1 cl && hterm_in_list ht2 cl
    in
  let rec hc_defined_in_list hc cl =
    match hc with
      | HBoolean(b) -> true
      | HNeg(n, hc1) -> hc_defined_in_list hc1 cl
      | HAtom(c, ht1, ht2) -> hterm_in_list ht1 cl && hterm_in_list ht2 cl
      | HDAtom(c, dt1, dt2) -> true
      | HOp(op, hc1, hc2) -> hc_defined_in_list hc1 cl && hc_defined_in_list hc2 cl
    in
  let rec search_impl lc ld hconj cl = (* Create a tuple (list of hConj without disjunction , list of tuple with each expression without neg operator) = (hConj list, [(h1,h2);...]) *)
    match hconj with
      | [] -> raise(Invalid_argument "search_impl: none hybridConjunction")
      | h :: [] -> 
        if (find_disj h true) 
        then (
          if (hc_defined_in_list h cl)
          then (lc, ([replace_neg h] @ ld))
          else (lc, ld)) 
        else (([h] @ lc), ld)
      (* | h :: [] -> if (find_disj h true) then (lc, ([h] @ ld)) else (([h] @ lc), ld) *)
      | h :: t -> 
        if (find_disj h true) 
        then (
          if (hc_defined_in_list h cl)
          then (search_impl lc ([replace_neg h] @ ld) t cl)
          else (search_impl lc ld t cl))
        else (search_impl ([h] @ lc) ld t cl)
      (* | h :: t -> if (find_disj h true) then (search_impl lc ([h] @ ld) t) else (search_impl ([h] @ lc) ld t) *)
    in
  let create_tuple s cl = (* Create a tuple (valuation, (list of hConj without disjunction, list of disjunction) ) *)
    match s with
      | (state, lval, hconj) -> (lval, (search_impl [] [] hconj cl))
    in
  let rec select_each_sval s cl lr = (* Select each element of the list *)
    match s with
      | [] -> raise(Invalid_argument "select_each_sval: not a super_valuation type or empty")
      | h :: [] -> [create_tuple h cl] @ lr
      | h :: t -> select_each_sval t cl ([create_tuple h cl] @ lr)
    in select_each_sval sval cl [];;

(* #trace separate_disj;;
separate_disj simpl_wp;; *)

let print_hConj h =
  let rec print_all_hConj l =
    match l with
      | [] -> ""
      | h::[] -> string_of_hybridConj (fst (snd h))
      | h::t -> string_of_hybridConj (fst (snd h)) ^ "\n" ^ print_all_hConj t
    in print_all_hConj h
;;

open Display;;
let print_hD h =
  let rec print_elt_hD l =
    match l with
      | [] -> ""
      | h::[] -> string_of_hybridCondition (h)
      | h::t -> string_of_hybridCondition (h) ^ "\n" ^ print_elt_hD t
    in
  let rec print_all_hD l =
    match l with
      | [] -> ""
      | h :: [] -> print_elt_hD h
      | h :: t -> print_elt_hD h ^ "\n\n" ^ print_all_hD t
    in print_all_hD h
;;

let print_hDisjl sval =
  let get_tuple s =
    match s with
      | (lval, cstr) -> string_of_hybridConj (snd cstr)
    in
  let rec eval_all_element s =
    match s with
      | [] -> raise(Invalid_argument "eval_all_element: list sval empty")
      | h :: [] -> get_tuple h
      | h :: t -> (get_tuple h) ^ "\n\n" ^ (eval_all_element t)
    in eval_all_element sval;;

let print_hDisjl_with_simplify_cstr_touch_delay sval =
  let rec get_each_elt l =
    match l with
      | [] -> ""
      | h::[] -> string_of_hybridCondition (fst h) ^ " ∨ " ^ string_of_hybridCondition (snd h) 
      | h::t -> string_of_hybridCondition (fst h) ^ " ∨ " ^ string_of_hybridCondition (snd h) ^ "\n" ^ get_each_elt t
    in
  let get_tuple s =
    match s with
      | (lval, cstr) -> get_each_elt (snd cstr)
    in
  let rec eval_all_element s =
    match s with
      | [] -> raise(Invalid_argument "eval_all_element: list sval empty")
      | h :: [] -> get_tuple h
      | h :: t -> (get_tuple h) ^ "\n\n" ^ (eval_all_element t)
    in eval_all_element sval;;

let print_hConjl sval =
  let get_tuple s =
    match s with
      | (lval, cstr) -> string_of_hybridConj (fst cstr)
    in
  let rec eval_all_element s =
    match s with
      | [] -> raise(Invalid_argument "eval_all_element: list sval empty")
      | h :: [] -> get_tuple h
      | h :: t -> (get_tuple h) ^ "\n\n" ^ (eval_all_element t)
    in eval_all_element sval;;


(* The function below determines if some constraints of celerities (included in "disjunctive formula") exist. 
We search this constraint in the hybridConjunction list. If it doesn't exist, we look for the neighbor 
celerity and assess the result for the celerity of the "disjunctive formula". If we can't have some 
information about the constraint of the celerity, we let the formula in the "disjunctive formula" list *)

let print_hdisjl_replaced hl =
  let rec c l =
    match l with
      | [] -> ""
      | h::[] -> (string_of_hybridCondition h)
      | h::t -> (string_of_hybridCondition h) ^ "\n" ^ (c t)
    in
  let rec b l =
    match l with
      | [] -> ""
      | h::t -> (c h) ^ "\n\n" ^ (b t)
    in
  let rec a l =
    match l with
      | [] -> ""
      | h::t -> (b h) ^ "\n" ^ (a t)
    in a hl;;

let print_hdisjl_final hd =
  let rec b l =
    match l with
      | [] -> ""
      | h::[] -> (string_of_hybridCondition h)
      | h::t -> (string_of_hybridCondition h) ^ "\n" ^ (b t)
    in
  let rec a l =
    match l with
      | [] -> ""
      | h::t -> (b h) ^ "\n" ^ (a t)
    in a hd;; 


let simplify_cstr_pi sval =
  let rec in_lval pi fl lval =
    match lval with
      | [] -> raise (Invalid_argument "in_lval : lval should not be empty")
      | h :: [] -> (match h with
              | (ht1,f) -> (ht1 = pi) && (f = fl))
      | h :: t -> (match h with
              | (ht1,f) -> ((ht1 = pi) && (f = fl)) || (in_lval pi fl t))      
    in
  let find_pi ht lval =
    match ht with
      | HAtom(comp, PiExit(v,i), HConstant(fl)) -> in_lval (PiExit(v,i)) fl lval
      | other -> raise(Invalid_argument "find_pi : should be 0 < PiExit < 1")
    in
  let find_cstr hcl h lc ld lval = 
  (* Main function which assesses if there exists a constraint of the current celerity in the hybridCondition list hcl, 
  else if there exists a constraint for its neighbor celerity, and put/remove/keep constraint in hybridCondition list 
  with or without "disjunctive formula" consequently *)
    match h with
      | HOp(op, HAtom(comp, PiExit(v,i), HConstant(fl)), HOp(And, ht1, ht2)) -> if (in_lval (PiExit(v,i)) fl lval) then ([ht2] @ ([ht1] @ lc),ld) else (lc,ld)
      | HOp(op, HOp(operator, ht1, ht2), h2)-> if ((find_pi ht1 lval) || (find_pi ht2 lval)) then (lc, ld) else ([h2] @ lc,ld)
      | HOp(op, h1, h2) ->  (lc, ([HOp(op, h1, h2)] @ ld))
      (* | HOp(op, h1, h2) ->  (lc, ([(h1,h2)] @ ld)) *)
      | other -> raise(Invalid_argument "find_cstr: Error")
    in
  let rec hd_list hcl hdl lc ld lval = (* Assumes all element in the list *)
    match hdl with
      | [] -> raise(Invalid_argument "hd_list: Error")
      | h :: [] -> ((fst(find_cstr hcl h lc ld lval)) @ hcl, snd(find_cstr hcl h lc ld lval))
      | h :: t -> hd_list hcl t (fst (find_cstr hcl h lc ld lval)) (snd (find_cstr hcl h lc ld lval)) lval
    in
  let get_tuple s =
    match s with
      | (lval, cstr) -> (lval, (hd_list (fst cstr) (snd cstr) [] [] lval))
    in
  let rec eval_all_element s l =
    match s with
      | [] -> raise(Invalid_argument "eval_all_element: list sval empty")
      | h :: [] -> [get_tuple h] @ l
      | h :: t -> eval_all_element t ([get_tuple h] @ l)
    in eval_all_element sval [];;
(* #trace simplify_cstr;; *)



let simplify_cstr_cel sval =
  let same_comp_neighbor c1 c2 = (* Compare between the celerity and its neighbor both comparators *)
      if (((c1 = GE) && (c2 = GT)) || 
          ((c1 = GE) && (c2 = GE)) || 
          ((c1 = LE) && (c2 = LT)) || 
          ((c1 = LE) && (c2 = LE)))
      then ("exist")
    else if (((c1 = GT) && (c2 = LT)) || 
             ((c1 = GT) && (c2 = LE)) || 
             ((c1 = LT) && (c2 = GT)) || 
             ((c1 = LT) && (c2 = GE)))
    then ("opposite")
  else (
    if ( (c1 = NEq) && ((c2 = GT) || (c2 = LT) || (c2 = NEq))) 
    then ("exist") 
    else if ((c1 = NEq) && (c2 = Eq))
      then ("opposite")
      else ("notexist"))
    in
  let same_comp c1 c2 = (* Compare between two formula of the same celerity their comparators. *)
      if (((c1 = GT) && (c2 = GT)) || 
          ((c1 = LT) && (c2 = LT)) || 
          ((c1 = GE) && (c2 = GT)) || 
          ((c1 = GE) && (c2 = GE)) || 
          ((c1 = LE) && (c2 = LT)) || 
          ((c1 = LE) && (c2 = LE)))
      then ("exist")
    else if (((c1 = GT) && (c2 = LT)) || 
             ((c1 = GT) && (c2 = LE)) || 
             ((c1 = LT) && (c2 = GT)) || 
             ((c1 = LT) && (c2 = GE)) || 
             ((c1 = GE) && (c2 = LT)) || 
             ((c1 = LE) && (c2 = GT)))
    then ("opposite")
  else (("notexist"))
    in
  let rec find_neighbor_in_lval lval cel c1 = (* Look for the neighbor celerity cel of the current one in the valulation list lval. c1 is the comparator of the "disjunctive formula" of the current celerity *)
    match lval with
      | [] -> raise(Invalid_argument "find_neighbor_in_lval: lval list empty")
      | h :: [] -> (match h with
        | (ht, f) -> if (ht = cel) 
            then (if (f > 0.) 
                          then (same_comp_neighbor c1 GE)
                        else (if f < 0.
                          then (same_comp_neighbor c1 LE) else ("notexist")))
            else "notexist")
      | h :: t -> (match h with
        | (ht, f) -> if (ht = cel) 
            then (if (f > 0.)
                          then (same_comp_neighbor c1 GE)
                        else (if f < 0.
                          then (same_comp_neighbor c1 LE) else ("notexist")))
            else (find_neighbor_in_lval t cel c1))
    in
  let rec find_cstr_neighbor hcl cel c1 = 
    (* Find a constraint of the Celerity cel in the hybridConjonction list,
     and compare its comparator with the comparator c1 *)
    match hcl with 
      | [] -> raise(Invalid_argument "find_cstr_neighbor: hybridConjunction list empty")
      | h :: [] -> (match h with
        | HAtom(c2, ht1, ht2) -> if (ht1 = cel) then (same_comp_neighbor c1 c2) else ("notexist")
        | _ -> raise(Invalid_argument "find_cstr_neighbor: error"))
      | h :: t -> (match h with
        | HAtom(c2, ht1, ht2) -> if (ht1 = cel) then (same_comp_neighbor c1 c2) else (find_cstr_neighbor t cel c1)
        | _ -> raise(Invalid_argument "find_cstr_neighbor: error"))
    in
  let search_neighbor h = (* Determine the neighbor celerity *)
    match h with
      | HAtom(comp, ht1, ht2) -> (match ht1 with
              | Celerity(v, w, n) -> if (n=0) then (Celerity(v, w, (n+1)), comp) else (Celerity(v, w, (n-1)), comp)
              | _ -> raise(Invalid_argument "search_neighbor: only Celerity compared with 0"))
      | _ -> raise(Invalid_argument "search_neighbor: Error")
    in
  let rec get_hterm h = (* Give the hybridCondition ht1 of the formula *)
    match h with 
      | HAtom(comp, ht1, ht2) -> ht1
      | _ -> raise(Invalid_argument "get_hterm: Error")
    in
  let rec get_comp h = (* Give the comparator of the hybridCondition HAtom*)
    match h with
      | HAtom(comp, ht1, ht2) -> comp
      | _ -> raise(Invalid_argument "get_comp: Error")
    in
  let rec exist_cstr ht1 c1 hcl = (* Give the result of both constraints of there exists a constraint of the current celerity in the hybridCondition list hcl *)
    match hcl with
    | [] -> raise(Invalid_argument "exist_cstr: hybridConjunction list empty")
    | hd :: [] -> if ((get_hterm hd) = ht1) then (same_comp c1 (get_comp hd)) else ("notexist")
    | hd :: tl -> if ((get_hterm hd) = ht1) then (same_comp c1 (get_comp hd)) else (exist_cstr ht1 c1 tl)
    in
  let rec find_all_cstr hcl h lc ld lval h1 op h2 = 
    match h with
      | HAtom(comp, ht1, ht2) -> if ((exist_cstr ht1 comp hcl)) = "exist" then (lc, ld)
                          else if ((exist_cstr ht1 comp hcl) = "opposite") then ([h1] @ lc, ld)
                        else if (find_cstr_neighbor hcl (fst (search_neighbor (HAtom(comp, ht1, ht2)))) (snd (search_neighbor (HAtom(comp, ht1, ht2))))) = "exist"
                        then (lc, ld)
                      else if (find_cstr_neighbor hcl (fst (search_neighbor (HAtom(comp, ht1, ht2)))) (snd (search_neighbor (HAtom(comp, ht1, ht2))))) = "opposite"
                      then ([h1] @ lc, ld)
                    else if (find_neighbor_in_lval lval (fst (search_neighbor (HAtom(comp, ht1, ht2)))) (snd (search_neighbor (HAtom(comp, ht1, ht2))))) = "exist"
                    then (lc, ld)
                  else if (find_neighbor_in_lval lval (fst (search_neighbor (HAtom(comp, ht1, ht2)))) (snd (search_neighbor (HAtom(comp, ht1, ht2))))) = "opposite"
                  then ([h1] @ lc, ld)
                else (lc, ([HOp(op, h1, h2)] @ ld))
      | HOp(op, ht1, ht2) -> if ((exist_cstr (get_hterm ht1) (get_comp ht1) hcl)) = "exist" then (lc, ld)
                          else if ((exist_cstr (get_hterm ht1) (get_comp ht1) hcl) = "opposite") then ([h1] @ lc, ld)
                        else if (find_cstr_neighbor hcl (fst (search_neighbor ht1)) (snd (search_neighbor ht1))) = "exist"
                        then (lc, ld)
                      else if (find_cstr_neighbor hcl (fst (search_neighbor ht1)) (snd (search_neighbor ht1))) = "opposite"
                      then ([h1] @ lc, ld)
                    else if (find_neighbor_in_lval lval (fst (search_neighbor ht1)) (snd (search_neighbor ht1))) = "exist"
                    then (lc, ld)
                  else if (find_neighbor_in_lval lval (fst (search_neighbor ht1)) (snd (search_neighbor ht1))) = "opposite"
                  then ([h1] @ lc, ld)
                else find_all_cstr hcl ht2 lc ld lval h1 op h2
      | _ -> raise(Invalid_argument "find_all_cstr: Should be of the form C_{v,w,k<n} > 0 ∧ C_{v,w,k>n} < 0")
    in
  let find_cstr hcl h lc ld lval = 
  (* Main function which assesses if there exists a constraint of the current celerity in the hybridCondition list hcl, 
  else if there exists a constraint for its neighbor celerity, and put/remove/keep constraint in hybridCondition list 
  with or without "disjunctive formula" consequently *)
    match h with
      | HOp(op, h1, h2) -> if ((exist_cstr (get_hterm h1) (get_comp h1) hcl)) = "exist" then (lc, ld)
                          else if ((exist_cstr (get_hterm h1) (get_comp h1) hcl) = "opposite") then ([h2] @ lc, ld)
                        else if (find_cstr_neighbor hcl (fst (search_neighbor h1)) (snd (search_neighbor h1))) = "exist"
                        then (lc, ld)
                      else if (find_cstr_neighbor hcl (fst (search_neighbor h1)) (snd (search_neighbor h1))) = "opposite"
                      then ([h2] @ lc, ld)
                    else if (find_neighbor_in_lval lval (fst (search_neighbor h1)) (snd (search_neighbor h1))) = "exist"
                    then (lc, ld)
                  else if (find_neighbor_in_lval lval (fst (search_neighbor h1)) (snd (search_neighbor h1))) = "opposite"
                  then ([h2] @ lc, ld)
                else if ((get_comp h1) = NEq) 
                  then (find_all_cstr hcl h2 lc ld lval h1 op h2)
                  else (lc, ([HOp(op, h1, h2)] @ ld))
                (* else (lc, ([(h1,h2)] @ ld)) *)
                (* else (find_cstr_on_touch_delay_comp hcl lc ld h1 h2)  *)
      | other -> raise(Invalid_argument "find_cstr: Error")
    in
  let rec hd_list hcl hdl lc ld lval = (* Assumes all element in the list *)
    match hdl with
      | [] -> raise(Invalid_argument "hd_list: Error")
      | h :: [] -> ((fst(find_cstr hcl h lc ld lval)) @ hcl, snd(find_cstr hcl h lc ld lval))
      | h :: t -> hd_list hcl t (fst (find_cstr hcl h lc ld lval)) (snd (find_cstr hcl h lc ld lval)) lval
    in
  let get_tuple s =
    match s with
      | (lval, cstr) -> (lval, (hd_list (fst cstr) (snd cstr) [] [] lval))
    in
  let rec eval_all_element s l =
    match s with
      | [] -> raise(Invalid_argument "eval_all_element: list sval empty")
      | h :: [] -> [get_tuple h] @ l
      | h :: t -> eval_all_element t ([get_tuple h] @ l)
    in eval_all_element sval [];;
(* #trace simplify_cstr;; *)


let simplify_cstr_touch_delay sval =
  let exist_cstr_touch_delay c1 c2 =
    if (((c1 = GT) && (c2 = GT)) ||
        ((c1 = LT) && (c2 = LT)) ||
        ((c1 = GE) && (c2 = GT)) ||
        ((c1 = GE) && (c2 = GE)) ||
        ((c1 = GE) && (c2 = Eq)) ||
        ((c1 = LE) && (c2 = LT)) ||
        ((c1 = LE) && (c2 = LE)) ||
        ((c1 = LE) && (c2 = Eq)) ||
        ((c1 = Eq) && (c2 = Eq)) ||
        ((c1 = NEq) && (c2 = GT)) ||
        ((c1 = NEq) && (c2 = LT)) ||
        ((c1 = NEq) && (c2 = NEq)))
    then ("exist")
    else (
      if (((c1 = GT) && (c2 = LT)) ||
          ((c1 = GT) && (c2 = LE)) ||
          ((c1 = GT) && (c2 = Eq)) ||
          ((c1 = LT) && (c2 = GT)) ||
          ((c1 = LT) && (c2 = GE)) ||
          ((c1 = LT) && (c2 = Eq)) ||
          ((c1 = GE) && (c2 = LT)) ||
          ((c1 = LE) && (c2 = GT)) ||
          ((c1 = Eq) && (c2 = GT)) ||
          ((c1 = Eq) && (c2 = LT)) ||
          ((c1 = Eq) && (c2 = NEq)) ||
          ((c1 = NEq) && (c2 = Eq)))
      then ("opposite")
      else ("notexist"))
    in
  let rec get_hterms h = (* Give the hybridCondition ht1 of the formula *)
    match h with 
      | HAtom(comp, ht1, ht2) -> (ht1,ht2)
      | _ -> raise(Invalid_argument "get_hterms: Error")
    in
  let rec get_comp h = (* Give the comparator of the hybridCondition HAtom*)
    match h with
      | HAtom(comp, ht1, ht2) -> comp
      | _ -> raise(Invalid_argument "get_comp: Error")
    in
  let rec assess_formula hd h2 = (* la fonction gère les formules C_{v,w,k<n} > 0 ∧ C_{v,w,k>n} < 0 
    et PiEntrance(v,i) comp PiExit(v,i) - C_{v,w,n} *) 
    match h2 with
      | HAtom(comp, ht1, ht2) -> 
        if (fst (get_hterms hd) = ht1) && (snd (get_hterms hd) = ht2)
        then (exist_cstr_touch_delay (get_comp h2) (get_comp hd))
        else ("notexist")
      | HOp(op, h1, h2) -> 
        if (fst (get_hterms hd) = fst (get_hterms h1)) && (snd (get_hterms hd) = snd (get_hterms h1))
        then (exist_cstr_touch_delay (get_comp h1) (get_comp hd))
        else assess_formula hd h1
      | _ -> raise(Invalid_argument "assess_formula: should be of the form C_{v,w,k<n} > 0 ∧ C_{v,w,k>n} < 0, or 
        of the form PiEntrance(v,i) comp PiExit(v,i) - C_{v,w,n}")
    in
  let rec find_cstr_on_touch_delay_comp hcl lc ld h1 h2 =
    match hcl with
      | [] -> (lc, ([(h1,h2)] @ ld))
      | h :: [] -> if (assess_formula h h2 = "exist")
                   then ((lc, ld))
                   else (
                    if (assess_formula h h2 = "opposite")
                    then (([h2] @ lc, ld))
                    else ((lc, ([(h1,h2)] @ ld))))
      | h :: t -> if (assess_formula h h2 = "exist")
                   then ((lc, ld))
                   else (
                    if (assess_formula h h2 = "opposite")
                    then (([h2] @ lc, ld))
                    else (find_cstr_on_touch_delay_comp t lc ld h1 h2))
    in
  let find_cstr hcl h lc ld lval = 
    match h with
      | HOp(op, h1, h2) -> find_cstr_on_touch_delay_comp hcl lc ld h1 h2 
      | other -> raise(Invalid_argument "find_cstr: Error")
    in
  let rec hd_list hcl hdl lc ld lval = (* Assumes all element in the list *)
    match hdl with
      | [] -> raise(Invalid_argument "hd_list: Error")
      | h :: [] -> ((fst(find_cstr hcl h lc ld lval)) @ hcl, snd(find_cstr hcl h lc ld lval))
      | h :: t -> hd_list hcl t (fst (find_cstr hcl h lc ld lval)) (snd (find_cstr hcl h lc ld lval)) lval
    in
  let get_tuple s =
    match s with
      | (lval, cstr) -> (lval, (hd_list (fst cstr) (snd cstr) [] [] lval))
    in
  let rec eval_all_element s l =
    match s with
      | [] -> raise(Invalid_argument "eval_all_element: list sval empty")
      | h :: [] -> [get_tuple h] @ l
      | h :: t -> eval_all_element t ([get_tuple h] @ l)
    in eval_all_element sval [];;
(* #trace simplify_cstr;; *)

(* Fonctions de test d'affichage à retirer quand fini *)
let val_of_simpl_wp w =
  match w with
    | [] -> raise(Invalid_argument "bouh")
    | h :: [] -> snd (snd h)
    | h :: t -> snd (snd h);;
(* #trace val_of_simpl_wp ;;
val_of_simpl_wp (separate_disj simpl_wp);; *)

let print_final r text =
  (fun (w) -> print_endline ("Voici le résultat de '" ^ text ^ "' :\n" ^ 
    (string_of_hybridConj w)))
    r ;;

let rec print_other_final r text l =
  match r with 
    | [] -> raise(Invalid_argument "nope")
    | h::[] -> print_final ([(fst h)] @ [(snd h)] @ l) text
    | h::t -> print_other_final t text ([(fst h)] @ [(snd h)] @ l);;

(* print_final (val_of_simpl_wp (separate_disj simpl_wp)) "hdisj";; *)
(* print_other_final (val_of_simpl_wp (dnf_result)) "hdisj with simplification" [];; *)




(********************)
(*** Simplify DNF ***)
(********************)

let select_hdisj s = 
  let rec select_hdisjl sval =
    match sval with
      | [] -> raise(Invalid_argument "select_hdisjl: super valuation list empty")
      | h::[] -> (match h with 
                        | (lval, (hConj, hDisj)) -> [hDisj])
      | h::t -> (match h with 
                        | (lval, (hConj, hDisj)) -> [hDisj] @ (select_hdisjl t))
    in select_hdisjl s ;;

let print_select_hdisj s =
  let rec display_tuple l =
    match l with
      | [] -> "No tuple there"
      | h :: [] -> string_of_hybridCondition (fst h) ^ " OR " ^ string_of_hybridCondition (snd h) ^ "\nEnd of tuples in select_hdisj"
      | h :: t -> string_of_hybridCondition (fst h) ^ " OR " ^ string_of_hybridCondition (snd h) ^ "\n" ^ (display_tuple t)
    in
  let rec display_list hd =
    match hd with
      | [] -> ""
      | h :: [] -> display_tuple h ^ "\n\nNo other disj list."
      | h :: t -> display_tuple h ^ "\nNew disj list: \n" ^ (display_list t)
    in display_list s;;


let replace_HOp_in_hdisjl hl =
  let rec replace_HOp h =
    match h with
      | HOp(op, h1, h2) -> (replace_HOp h1) @ (replace_HOp h2)
      | other_hterm -> [other_hterm]
    in
  let rec select_each_tuple l =
    match l with
      | [] -> []
      | h :: [] -> [(replace_HOp (fst h), replace_HOp (snd h))]
      | h :: t -> [(replace_HOp (fst h), replace_HOp (snd h))] @ select_each_tuple t
    in
  let rec select_each_list l =
    match l with
      | [] -> []
      | h :: [] -> [select_each_tuple h]
      | h :: t -> [select_each_tuple h] @ (select_each_list t)
  in select_each_list hl;;


let print_replace_HOp_in_hdisjl hl =
  let rec display_each_tuple l =
    match l with
      | [] -> "No elt in tuple"
      | h :: [] -> string_of_hybridConj (fst h) ^ " OR " ^ string_of_hybridConj (snd h) ^ "\nEnd of tuples in replacement\n"
      | h :: t -> string_of_hybridConj (fst h) ^ " OR " ^ string_of_hybridConj (snd h) ^ "\n" ^ (display_each_tuple t)
    in
  let rec display_each_list l =
    match l with
      | [] -> "No disj list"
      | h :: [] -> display_each_tuple h ^ "\nNo other disj list"
      | h :: t -> display_each_tuple h ^ "\nNew disj list:\n" ^ (display_each_list t)
    in display_each_list hl;;



(* The function below combines all formulas in disjunction to obtain a list with all
possibilities. The number of elements in the output list corresponds to the number of 
expressions in disjunction, and so the number of files to create for constraint solver 
REMARQUE : On ressort une ((hdisj list) list car on peut avoir plusieurs solutions *)
let comb_elt_in_disj hl =
  let assess_comp_neighbor c1 c2 =
    if (((c1 = GT) && (c2 = LT)) || 
        ((c1 = LT) && (c2 = GT)))
    then (true)
    else (false)
    in
  let assess_comp c1 c2 =
    if (((c1 = GT) && (c2 = LT)) || 
        ((c1 = GT) && (c2 = LE)) || 
        ((c1 = GT) && (c2 = Eq)) || 
        ((c1 = LT) && (c2 = GT)) || 
        ((c1 = LT) && (c2 = GE)) || 
        ((c1 = LT) && (c2 = Eq)) || 
        ((c1 = GE) && (c2 = LT)) || 
        ((c1 = LE) && (c2 = GT)) || 
        ((c1 = Eq) && (c2 = GT)) || 
        ((c1 = Eq) && (c2 = LT)) || 
        ((c1 = Eq) && (c2 = NEq)) || 
        ((c1 = NEq) && (c2 = Eq)))
    then (true)
    else (false)
    in
  let rec hatom_are_opposite elt1 elt2 =
    match (elt1, elt2) with
      | (HAtom(c1, Celerity(v1,w1,n), fl1), HAtom(c2, Celerity(v2,w2,m), fl2)) -> 
        (if ((v1 = v2) && (w1 = w2) && (fl1 = fl2))
         then (
          if (m = n)
          then (assess_comp c1 c2) 
          else (
            if ((m = n-1) || (m = n+1)) 
            then (assess_comp_neighbor c1 c2) 
            else (false))) 
         else (false))
      | _ -> false
    in
  let rec elt_and_list_are_opposite elt l =
    match l with
      | [] -> false
      | h :: [] -> hatom_are_opposite elt h
      | h :: t -> (hatom_are_opposite elt h) || (elt_and_list_are_opposite elt t)
    in
  let rec are_lists_incompatible elts l =
    match elts with
      | [] -> false
      | h :: [] -> elt_and_list_are_opposite h l
      | h :: t -> (elt_and_list_are_opposite h l) || (are_lists_incompatible t l)
    in
  let rec in_list elt l =
    match l with
      | [] -> false
      | h::[] -> if (elt = h) then true else false
      | h :: t -> if (elt = h) then true else (in_list elt t)
    in
  let simplify_comps comp1 comp2 =
    match (comp1, comp2) with
      | (GT,GT) | (GT,GE) | (GT,NEq) | (GE,GT) | (GE,NEq) | (NEq,GT) | (NEq,GE) -> GT
      | (LT,LT) | (LT,LE) | (LT,NEq) | (LE,LT) | (LE,NEq) | (NEq,LT) | (NEq,LE) -> LT
      | (GE,GE) -> GE
      | (LE,LE) -> LE
      | (GE,LE) | (GE,Eq) | (LE,GE) | (LE,Eq) | (Eq,GE) | (Eq,LE) | (Eq,Eq) -> Eq
      | (NEq,NEq) -> NEq
      | other -> raise(Invalid_argument "simplify_comps: Hatoms should not be the same or not simplified before")
    in
  let rec simplify_hatom elt l =
    match l with
      | [] -> elt
      | h :: t -> 
        match (elt,h) with
          | (HAtom(comp1, ht1, fl1), HAtom(comp2, ht2, fl2)) ->
            if ((ht1 = ht2) && (fl1 = fl2))
            then (HAtom((simplify_comps comp1 comp2), ht1, fl1))
            else (simplify_hatom elt t)
          | _ -> raise(Invalid_argument "simplify_hatom: should be of the form C_{v,w,n} comp float")
    in
  let rec hatom_in_list elt l =
    match l with
      | [] -> false
      | h :: t -> 
        match (elt,h) with
          | (HAtom(comp1, ht1, fl1), HAtom(comp2, ht2, fl2)) -> 
            if (ht1 = ht2 && fl1 = fl2)
            then (true)
            else (hatom_in_list elt t)
          | (_,_) -> raise(Invalid_argument "hatom_in_list: should be HAtom")
    in
  let rec simplify_combination elts ld lf =
    match elts with
      | [] -> lf
      | h :: t -> 
        let complete_list_without_current_elt = List.filter (fun x -> x <> h) ld in
        if (hatom_in_list h complete_list_without_current_elt)
        then (let new_hatom = simplify_hatom h complete_list_without_current_elt in
          if (hatom_in_list new_hatom lf)
          then (simplify_combination t ld lf)
          else (simplify_combination t ld ([new_hatom] @ lf)))
        else (simplify_combination t ld ([h] @ lf))
    in
  let rec concat_simplify l1 l2 =
    match l1 with
      | [] -> l2
      | h :: t -> 
        if (in_list h l2)
        then (concat_simplify t l2)
        else (concat_simplify t ([h] @ l2))
    in
  let rec find_neighbor elt l =
    match l with
      | [] -> [elt]
      | h :: t -> 
        match (elt,h) with
          | (HAtom(comp1, Celerity(v1,w1,n1), fl1),HAtom(comp2, Celerity(v2,w2,n2), fl2)) ->
            if ((v1 = v2) && 
                (w1 = w2) && 
                (fl1 = fl2) && 
                ((n1 = n2+1) || (n1 = n2-1)) &&
                (((comp1 = NEq) && (comp2 = NEq)) ||
                  ((comp1 = NEq) && (comp2 = GT)) ||
                  ((comp1 = NEq) && (comp2 = LT)) ||
                  ((comp1 = GT) && (comp2 = NEq)) ||
                  ((comp1 = LT) && (comp2 = NEq))))
            then ([HAtom(GT, HTermRelation(Times, Celerity(v2, w2, n2), Celerity(v1, w1, n1)), HConstant(0.))])
            else (find_neighbor elt t)
          | (_, _) -> find_neighbor elt t
    in
  let rec elt_in_list elt l =
    match l with
      | [] -> false
      | h :: t ->
        match (elt,h) with
          | (HAtom(c1, Celerity(v1,w1,n1), fl1),HAtom(c2, Celerity(v2,w2,n2), fl2)) ->
            if ((c1 = c2) && (v1 = v2) && (w1 = w2) && (n1 = n2) && (fl1 = fl2))
            then (true)
            else (elt_in_list elt t)
          | (HAtom(
              comp1, 
              Celerity(v1,w1,n1), 
              fl1),
             HAtom(
              GT,
              HTermRelation(
                Times,
                Celerity(v2,w2,n2),
                Celerity(v3,w3,n3)),
              HConstant(0.))) ->
            if (((v1 = v2) && (w1 = w2) && (n1 = n2)) ||
                ((v1 = v3) && (w1 = w3) && (n1 = n3)))
            then (true)
            else (elt_in_list elt t)
          | (_,_) -> elt_in_list elt t
    in
  let rec simplify_neighbors elts ld lf =
    match elts with
      | [] -> lf
      | h :: t -> 
        if (elt_in_list h lf) 
        then (simplify_neighbors t ld lf)
        else (simplify_neighbors t ld ((find_neighbor h t) @ lf)) 
    in
  let rec compare_lists elts l =
    if (are_lists_incompatible elts l) 
    then ([]) 
    else (
      let res = (concat_simplify elts l) in 
      let simpl_neighbor = simplify_neighbors res res [] in
      simplify_combination simpl_neighbor simpl_neighbor [])
    in
  let rec map_simplify f l =
    match l with
      | [] -> []
      | h::t -> let res = f h in
        if (res <> []) 
        then ([res] @ map_simplify f t) 
        else (map_simplify f t)
    in
  let rec combine_disj hd =
    match hd with
      | [] -> []
      | h :: [] -> [fst h] @ [snd h]
      | h :: t -> map_simplify (compare_lists (fst h)) (combine_disj t) @ map_simplify (compare_lists (snd h)) (combine_disj t)
    in
  let rec combine_disj_list hd =
    match hd with
      | [] -> raise(Invalid_argument "combine_disj_list: hdisj list empty")
      | h :: [] -> [combine_disj h]
      | h :: t -> [combine_disj h] @ (combine_disj_list t)
    in combine_disj_list hl;;


let print_comb_elt_in_disj hd =
  let rec display_each_sublist l =
    match l with
      | [] -> "No elt in sublist"
      | h :: [] -> string_of_hybridConj h ^ "\nEnd of tuples in comb_elt_in_disj\n"
      | h :: t -> string_of_hybridConj h ^ "\n" ^ display_each_sublist t
    in
  let rec display_each_list l =
    match l with
      | [] -> "No disj list"
      | h :: [] -> display_each_sublist h ^ "\nNo other disj list"
      | h :: t -> display_each_sublist h ^ "\nNew disj list:\n" ^ (display_each_list t)
    in display_each_list hd;;


(***********************************)
(*** Simplify Hybrid Conjunction ***)
(***********************************)

(* get_hconj function select hybridConjunction list to display into the output files *)
let get_hconj sval =
  let rec select_hconj s =
    match s with
      | (lval, (hCond,_)) -> hCond
    in
  let rec select_element s l =
    match s with
      | [] -> raise(Invalid_argument "select_element: lvaluation empty")
      | h::[] -> [select_hconj h] @ l
      | h::t -> select_element t ([select_hconj h] @ l)
    in
    select_element sval [];;


(* We replace all HOp() by the elements of the corresponding HAtom() in order to 
make the next simplification easier  *)
let replace_HOp_in_hconl hl =
  let rec replace_HOp h =
    match h with
      | HOp(op, h1, h2) -> (replace_HOp h1) @ (replace_HOp h2)
      | other_hterm -> [other_hterm]
    in
  let rec select_each_hterm l =
    match l with
      | [] -> raise(Invalid_argument "select_each_hterm: one hconj list empty")
      | h :: [] -> replace_HOp h
      | h :: t -> replace_HOp h @ (select_each_hterm t)
    in
  let rec select_each_elt l =
    match l with
      | [] -> raise(Invalid_argument "select_each_elt: all hconj lists empty")
      | h :: [] -> [select_each_hterm h]
      | h :: t -> [select_each_hterm h] @ (select_each_elt t)
  in select_each_elt hl;;


(* We simplify the hybrid conjunction list in removing the replicate of a formula,
or reducing formulas (A >= B && A != B ...)  *)
let simplify_hconj hconj =
  let rec in_list elt l =
    match l with
      | [] -> false
      | h :: t -> if (elt = h) then (true) else (in_list elt t)
    in
  let new_comp c1 c2 =
    if (((c1 = LT) && (c2 = LT)) ||
        ((c1 = LT) && (c2 = LE)) ||
        ((c1 = LT) && (c2 = NEq)) ||
        ((c1 = LE) && (c2 = LT)) ||
        ((c1 = LE) && (c2 = NEq)) ||
        ((c1 = NEq) && (c2 = LT)) ||
        ((c1 = NEq) && (c2 = LE)))
    then ((LT, -0.001))
    else (
        if ((c1 = GT) && (c2 = GT)) ||
            ((c1 = GT) && (c2 = GE)) ||
            ((c1 = GT) && (c2 = NEq)) ||
            ((c1 = GE) && (c2 = GT)) ||
            ((c1 = GE) && (c2 = NEq)) ||
            ((c1 = NEq) && (c2 = GT)) ||
            ((c1 = NEq) && (c2 = GE))
        then ((GT, 0.001))
        else (
          if ((c1 = LE) && (c2 = LE))
          then ((LE, 0.))
          else (
            if ((c1 = GE) && (c2 = GE))
            then ((GE, 0.))
            else (
              if ((c1 = LE) && (c2 = GE)) ||
                 ((c1 = LE) && (c2 = Eq)) ||
                 ((c1 = GE) && (c2 = LE)) ||
                 ((c1 = GE) && (c2 = Eq)) ||
                 ((c1 = Eq) && (c2 = LE)) ||
                 ((c1 = Eq) && (c2 = GE)) ||
                 ((c1 = Eq) && (c2 = Eq))            
              then ((Eq, 0.))
              else (
                if ((c1 = NEq) && (c2 = NEq))
                then ((NEq, 0.))
                else (raise (Invalid_argument "Opposite formulas found. Check your Hoare triple (path or Celerities defined in the Postcondition.)")))))))
  in
  let rec simplify_formula ha l lf =
    match l with
      | [] -> if (in_list ha lf) then ([]) else ([ha])
      | h :: t -> 
        match (ha,h) with
          | (HAtom(comp1, ht1_1, ht1_2), HAtom(comp2, ht2_1, ht2_2)) ->
            if (ht1_1 = ht2_1 && ht1_2 = ht2_2)
            then (
              let (comp, fl) = new_comp comp1 comp2 in
              let new_hatom = HAtom(comp, ht1_1, ht1_2) in
              if (in_list new_hatom lf) then ([]) else ([new_hatom]) )
            else (
              if (ht1_1 = ht2_1 &&
                  (
              (ht1_2 = HConstant(0.) && ht2_2 = HConstant(0.)) ||
              (ht1_2 = HConstant(0.) && ht2_2 = HConstant(0.001)) ||
              (ht1_2 = HConstant(0.) && ht2_2 = HConstant(-0.001)) ||
              (ht1_2 = HConstant(0.001) && ht2_2 = HConstant(0.)) ||
              (ht1_2 = HConstant(-0.001) && ht2_2 = HConstant(0.)) ||
              (ht1_2 = HConstant(0.001) && ht2_2 = HConstant(0.001)) ||
              (ht1_2 = HConstant(0.001) && ht2_2 = HConstant(-0.001)) ||
              (ht1_2 = HConstant(-0.001) && ht2_2 = HConstant(0.001)) ||
              (ht1_2 = HConstant(-0.001) && ht2_2 = HConstant(-0.001))))
              then (
                let (comp, fl) = new_comp comp1 comp2 in
                let new_hatom = HAtom(comp, ht1_1, HConstant(fl)) in
                if (in_list new_hatom lf) then ([]) else ([new_hatom]))
              else (simplify_formula ha t lf))
          | (_, _) -> simplify_formula ha t lf
    in
  let rec simplify_hconj_list l ld lf =
    match l with
      | [] -> lf 
      | h :: t ->
        let list_without_current_elt = List.filter (fun x -> x <> h) ld in
        simplify_hconj_list t ld ((simplify_formula h list_without_current_elt lf) @ lf)
    in
  let rec select_each_list hc =
    match hc with
      | [] -> raise(Invalid_argument "select_each_list: all hconj list empty")
      | h :: [] -> [simplify_hconj_list h h []]
      | h :: t -> [simplify_hconj_list h h []] @ (select_each_list t)
    in select_each_list hconj;;


let print_hConjl_simplified hconj =
  let rec print_hconj_list l =
    match l with
      | [] -> ""
      | h :: [] -> string_of_hybridCondition h ^ "\nEnd of hconj list"
      | h :: t -> string_of_hybridCondition h ^ "\n" ^ print_hconj_list t
    in
  let rec eval_each_element l =
    match l with
      | [] -> ""
      | h :: [] -> print_hconj_list h
      | h :: t -> print_hconj_list h ^ "\nNo other hconj list\n" ^ (eval_each_element t)
    in eval_each_element hconj;;
