(* This program reads input files containing hybrid Hoare logic *)
(* Starting from this input file, a parser is used in order to create 
the *)

open Types;;
open Printf;;
open Display;;

let isCyclic = ref false;;

(* Definition of the different types given by the parser *)
let eval_hoareTriple e =
	let eval_duration t =
		match t with
			| Parser.ValueTime(vt) -> ValueTime(float_of_string(vt))
			| Parser.VarTime(vt) -> VarTime(vt)
		in 
	let rec eval_symbol_slide s =
		match s with
			| '+' -> PlusSlide
			| '-' -> MinusSlide
			| _ -> raise(Invalid_argument "eval_symbol : Should have the '+', '-' symbol for Slide or none")
		in
	let eval_comp c =
		match c with
			| "<="|"=<" -> LE
			| "<" -> LT
			| ">="|"=>" -> GE
			| ">" -> GT
			| "=" -> Eq
			| "!=" -> NEq
			| _ -> raise(Invalid_argument "eval_comp: Need a comparator for the Celerities of assertions")
		in
	let rec eval_resources w = 
		match w with
			| Parser.Multiplex(m) -> [m]
			| Parser.SetMultiplexes(m,r) -> [m] @ eval_resources(r)
		in
	let eval_symbol_hterm s =
		match s with
			| '+' -> Plus
			| '-' -> Minus
			| '*' -> Times
			| '/' -> Divide
			| _ -> raise(Invalid_argument "eval_symbol_hterm : Should have +, -,* or / symbol in relation between Hybrid terms")
		in
	let rec eval_hybridTerm ht =
		match ht with
			| Parser.PiEntrance(p) -> PiEntrance(p,0)
			| Parser.PiExit(p) -> PiExit(p,0)
			| Parser.Celerity(v,w,n) -> Celerity(v,eval_resources(w),int_of_string(n))
			| Parser.CelerityNR(v,n) -> Celerity(v,[],int_of_string(n))
			| Parser.HTermRelation(ht1,connective,ht2) -> HTermRelation(eval_symbol_hterm(connective),eval_hybridTerm(ht1),eval_hybridTerm(ht2))
			| Parser.HConstant(cst) -> HConstant(float_of_string(cst))
			| Parser.HTime(t) -> HTime(eval_duration(t))
		in
	let rec eval_assertion ea =
		match ea with
			| Parser.ASlide(v)-> ASlide(EqSlide,v)
			| Parser.ASlideSymbol(s,v)-> ASlide(eval_symbol_slide(s),v)
			| Parser.ACelerity(v,comp,value)-> ACelerity(eval_comp(comp),v,eval_hybridTerm(value))
			| Parser.ABoolean(a)-> ABoolean(boolean_of_string(a))
			| Parser.AOpAnd(a1,a2)-> AOp(And,eval_assertion(a1),eval_assertion(a2))
			| Parser.AOpOr(a1,a2)-> AOp(Or,eval_assertion(a1),eval_assertion(a2))
			| Parser.ANeg(a)-> ANeg(Neg,eval_assertion(a))
			| Parser.APar(a)-> eval_assertion(a)
		in
	let eval_symbol s =
		match s with
			| '+' -> PlusSymbol
			| '-' -> MinusSymbol
			| _ -> raise(Invalid_argument "eval_symbol : Should have + or - symbol for the discrete path atom")
		in
	let eval_dpa d =
		match d with
			| Parser.DpaValue(v,s) -> (v,eval_symbol(s))
		in
	let eval_elementaryPath ep =
		match ep with
			| Parser.Tuple(d,at,dpa) -> Tuple(eval_duration(d),eval_assertion(at),eval_dpa(dpa))
		in
	let eval_symbol_dterm s =
		match s with
			| '+' -> DPlus
			| '-' -> DMinus
			| _ -> raise(Invalid_argument "eval_symbol_dterm : Should have + or - symbol in relation between Discrete terms")
		in
	let eval_discreteTerm d =
		match d with
			| Parser.Eta(v) -> Eta(v)
			| Parser.DConstant(v) -> DConstant(int_of_string(v))
		in
	let rec eval_discreteAtom d =
		match d with
			| Parser.DTermUn(dt) -> eval_discreteTerm(dt)
			| Parser.DTermRelation(dt1,connective,dt2) -> DTermRelation(eval_symbol_dterm(connective),eval_discreteAtom(dt1),eval_discreteAtom(dt2)) 
		in
	let rec eval_discreteCondition d =
		match d with
			| Parser.DBoolean(c1) -> DBoolean(boolean_of_string(c1))
			| Parser.DAtom(d1,comp,d2) -> DAtom(eval_comp(comp),eval_discreteAtom(d1),eval_discreteAtom(d2))
			| Parser.DOpAnd(d1,d2) -> DOp(And,eval_discreteCondition(d1),eval_discreteCondition(d2))
			| Parser.DOpOr(d1,d2) -> DOp(Or,eval_discreteCondition(d1),eval_discreteCondition(d2))
			| Parser.DNeg(d1) -> DNeg(Neg,eval_discreteCondition(d1))
			| Parser.DPar(d1) -> eval_discreteCondition(d1)
		in
	let rec eval_hybridCondition c =
		match c with
			| Parser.HBoolean(h1) -> HBoolean(boolean_of_string(h1))
			| Parser.HAtom(h1,comp,h2) -> HAtom(eval_comp(comp),eval_hybridTerm(h1),eval_hybridTerm(h2))
			| Parser.HDAtom(hd1,comp,hd2) -> HDAtom(eval_comp(comp),eval_discreteAtom(hd1),eval_discreteAtom(hd2))
			| Parser.HOpAnd(h1,h2) -> HOp(And,eval_hybridCondition(h1),eval_hybridCondition(h2))
			| Parser.HOpOr(h1,h2) -> HOp(Or,eval_hybridCondition(h1),eval_hybridCondition(h2))
			| Parser.HNeg(h1) -> HNeg(Neg,eval_hybridCondition(h1))
			| Parser.HPar(h1) -> eval_hybridCondition(h1)
		in
	let rec eval_path p =
		match p with
			| Parser.EPath(ep) -> eval_elementaryPath(ep)
			| Parser.Seq(ep,p) -> Seq(eval_elementaryPath(ep),(eval_path p))
		in
	let eval_post p =
		match p with
			| Parser.PostCondition(dc,hc) -> Postcondition(eval_discreteCondition(dc),eval_hybridCondition(hc))
		in
	let rec eval_multformula m =
		match m with
			| Parser.MAtom(v,comp,i) -> (match comp with
						| "=" -> MPropBin(And,MAtom(v,int_of_string(i)),MPropUn(Neg,MAtom(v,int_of_string(i)+1)))
						| "!=" -> MPropUn(Neg,MPropBin(And,MAtom(v,int_of_string(i)),MPropUn(Neg,MAtom(v,int_of_string(i)+1))))
						| ">" -> MAtom(v,int_of_string(i)+1)
						| ">="|"=>" -> MAtom(v,int_of_string(i))
						| "<" -> MPropUn(Neg,MAtom(v,int_of_string(i)))
						| "<="|"=<" -> MPropUn(Neg,MAtom(v,int_of_string(i)+1))
						| _ -> raise(Invalid_argument "eval_Mcomp : not a comparator !"))
			| Parser.MPropNeg(m1) -> MPropUn(Neg,eval_multformula(m1)) 
			| Parser.MPropAnd(m1,m2) -> MPropBin(And,eval_multformula(m1),eval_multformula(m2))
			| Parser.MPropOr(m1,m2) -> MPropBin(Or,eval_multformula(m1),eval_multformula(m2))
			| Parser.MPropPar(m1) -> eval_multformula(m1)
		in
	let eval_formula f =
		match f with
			| Parser.MFormula(f1) -> eval_multformula(f1)
		in
	let rec eval_varlist v =
		match v with
			| Parser.MVar(v1) -> [v1]
			| Parser.MVarList(v1,v2) -> [v1] @ eval_varlist(v2)
		in
	let rec var_as_target v lt =
	match lt with
		| [] -> raise(Invalid_argument "var_as_target: none target for each multiplex!")
		| h::[] -> if (v=h) then true else false
		| h::t -> if (v=h) then true else (var_as_target v t)
		in
	let rec create_list_target v m t =
		match v with
			| [] -> raise(Invalid_argument "create_list_target: none variable defined!")
			| (var,(n,l))::[] -> if (var_as_target var t) then [(var,(n, l @ [m]))] else [(var,(n, l))]
			| (var,(n,l))::tail -> if (var_as_target var t) then [(var,(n, l @ [m]))] @ create_list_target tail m t else [(var,(n, l))] @ create_list_target tail m t
		in
	let eval_target t =
		match t with
			| Parser.MTarget(t1) -> eval_varlist(t1)
		in
	let eval_predecessor p lv lm =
		match p with
			| Parser.VarDefinition(v,n) -> (lv @ [(v,(n,[]))] , lm)
			| Parser.MultDefinition(m,f,t) ->  (create_list_target lv m (eval_target t), lm @ [(m,eval_formula(f))])
		in
	let rec eval_influenceGraph s lv lm =
		match s with
			| Parser.PredecessorDef(p) -> eval_predecessor p lv lm
			| Parser.SeveralVarDef(p,s1) -> eval_influenceGraph s1 (fst (eval_predecessor p lv lm)) (snd (eval_predecessor p lv lm))
		in
	let identify_path e =
		match e with
			| Parser.HoareTriple(path,post) -> HoareTriple(eval_path(path),eval_post(post))
			| Parser.InfluenceGraph(s) -> InfluenceGraph(eval_influenceGraph(s) [] [])
			| Parser.IsCyclic(boolean) -> IsCyclic(boolean)
	in identify_path e
;;


let eval_duration t =
	match t with
		| ValueTime(vt) -> "ValueTime(" ^ string_of_float(vt) ^ ")"
		| VarTime(vt) -> "VarTime(\"" ^ vt ^ "\")"

let rec eval_hybridTerm ht =
	match ht with
	| PiEntrance(p,i) -> "PiEntrance(\"" ^ p ^ "\"," ^ string_of_int i ^ ")"
	| PiExit(p,i) -> "PiExit(\"" ^ p ^ "\"," ^ string_of_int i ^ ")"
	| Celerity(v,w,n) -> "Celerity(\"" ^ v ^ "\",[" ^ (string_of_list_mult string_of_mult ";"  w) ^ "]," ^ string_of_int n ^ ")"
	| HTermRelation(connective,ht1,ht2) -> "HTermRelation(" ^ string_of_termSymbol connective ^ "," ^ eval_hybridTerm(ht1) ^ "," ^ eval_hybridTerm(ht2) ^ ")"
	| HConstant(cst) -> "HConstant(" ^ string_of_float(cst) ^ ")"
	| HTime(t) -> "HTime(" ^ eval_duration(t) ^ ")"
;;

let rec eval_discreteTerm d =
	match d with
		| Eta(v) -> "Eta(\"" ^ v ^ "\")"
		| DConstant(v) -> "DConstant(" ^ string_of_int v ^ ")"
		| DTermRelation(connective,dt1,dt2) -> "DTermRelation(" ^ string_of_discreteTermSymbol(connective) ^ "," ^ eval_discreteTerm(dt1) ^ "," ^ eval_discreteTerm(dt2) ^ ")" 
;;




let path_to_string p =
	let rec eval_assertion ea =
		match ea with
			| ASlide(sy,v)-> "ASlide(" ^ string_of_slideSymbol sy ^ ",\"" ^ v ^ "\")"
			| ANoSlide(sy,v)-> "ANoSlide(" ^ string_of_slideSymbol sy ^ ",\"" ^ v ^ "\")"
			| ACelerity(c,v,value)-> "ACelerity(" ^ (string_of_comp c) ^ ",\"" ^ v ^ "\"," ^ eval_hybridTerm(value) ^ ")"
			| ABoolean(a)-> "ABoolean(" ^ string_of_boolean a ^ ")"
			| AOp(op,a1,a2)-> "AOp(" ^ string_of_operator op ^ "," ^ eval_assertion(a1) ^ "," ^ eval_assertion(a2) ^ ")"
			| ANeg(n,a)-> "ANeg(" ^ string_of_neg n ^ "," ^ eval_assertion(a) ^ ")"
		in
	let rec identify_path ep =
		match ep with
			| Tuple(d,at,dpa) -> "Tuple(" ^ eval_duration(d) ^ "," ^ eval_assertion(at) ^ "," ^ string_of_dpa(dpa) ^ ")"
			| Nothing -> "Nothing"
			| Seq(ep,p) -> "Seq(" ^ identify_path(ep) ^ "," ^ (identify_path p) ^ ")" 
		in identify_path p
;;

let rec dcondition_to_string d =
	match d with
		| DBoolean(c1) -> "DBoolean(" ^ string_of_boolean c1 ^ ")"
		| DAtom(comp,d1,d2) -> "DAtom(" ^ string_of_comp(comp) ^ "," ^ string_of_discreteTerm d1 ^ "," ^ string_of_discreteTerm  d2 ^ ")"
		| DOp(op,d1,d2) -> "DOp(" ^ string_of_operator op ^ "," ^ dcondition_to_string(d2) ^ "," ^ dcondition_to_string(d1) ^ ")"
		| DNeg(n,d1) -> "DNeg(" ^ string_of_neg n ^ "," ^ dcondition_to_string(d1) ^ ")"
;;

let rec hcondition_to_string h =
	match h with
		| HBoolean(h1) -> "HBoolean(" ^ string_of_boolean h1 ^ ")"
		| HAtom(comp,h1,h2) -> "HAtom(" ^ string_of_comp(comp) ^ "," ^ eval_hybridTerm(h1) ^ "," ^ eval_hybridTerm(h2) ^ ")"
		| HDAtom(comp,hd1,hd2) -> "HDAtom(" ^ string_of_comp(comp) ^ "," ^ eval_discreteTerm(hd1) ^ "," ^ eval_discreteTerm(hd2) ^ ")"
		| HOp(op,h1,h2) -> "HOp(" ^ string_of_operator op ^ "," ^ hcondition_to_string(h2) ^ "," ^ hcondition_to_string(h1) ^ ")"
		| HNeg(n,h1) -> "HNeg(" ^ string_of_neg n ^ hcondition_to_string(h1) ^ ")"
;;


let rec vars_to_string vl =
	match vl with
		| [] -> raise(Invalid_argument "vars_to_string: None variable defined !")
		| (v,(n,rl))::[] -> "(\"" ^ v ^ "\",(" ^ n ^ "," ^  (string_of_list_delim id_string ";" "[" "]" rl) ^ "))" 
		| (v,(n,rl))::t -> "(\"" ^ v ^ "\",(" ^ n ^ "," ^ (string_of_list_delim id_string ";" "[" "]" rl) ^ ")); " ^ (vars_to_string t)
;;  

let rec multformula_to_string mf =
	match mf with
	| MAtom(v,i) -> "MAtom(\"" ^ v ^ "\"," ^ string_of_int i ^ ")"
	| MPropUn(n,mf1) -> "MPropUn(" ^ string_of_neg n ^ "," ^ multformula_to_string mf1 ^ ")"
	| MPropBin(op,mf1,mf2) -> "MPropBin(" ^ string_of_operator op ^ "," ^ multformula_to_string mf1 ^ "," ^ multformula_to_string mf2 ^ ")"
;;

let rec mults_to_string ml =
	match ml with
		| [] -> raise(Invalid_argument "mults_to_string: None multiplex defined !")
		| (m,mf)::[] -> "(\"" ^ m ^ "\"," ^ multformula_to_string mf ^ ")"
		| (m,mf)::t -> "(\"" ^ m ^ "\"," ^ multformula_to_string mf ^ "); " ^ (mults_to_string t)
;; 

let print_data p =
	let display_path p =
		match p with
			| InfluenceGraph(ig) -> "let vars = [" ^ vars_to_string (fst ig) ^ "];;\nlet mults =[" ^ mults_to_string (snd ig) ^ "];;\n\n"
			| HoareTriple(path,post) -> (match post with
						| Postcondition(dc,hc) -> "let prog0 = " ^ path_to_string(path) ^ ";;\n\nlet d0 = " ^ dcondition_to_string (dc) ^ ";;\n\nlet h0 = ref (" ^ hcondition_to_string (hc) ^ ");;\n\n")
			| IsCyclic(boolean) -> "let isCyclic = " ^ string_of_bool(boolean) ^ ";;\n\n"
	in display_path p
;;


let write_data oc r =
	match r with
		| HoareTriple(path,post) -> fprintf oc "%s" (print_data (HoareTriple(path,post)))
		| InfluenceGraph(ig) -> fprintf oc "%s" (print_data (InfluenceGraph(ig)))
		| IsCyclic(b) -> fprintf oc "%s"  (print_data (IsCyclic(b)))
;;


let rec parse oe =
	let lexbuf = Lexing.from_channel (oe) in
	let n = ref 0 in
	try
		while true do
			let token = Parser.program Lexer.token lexbuf in
			let res = eval_hoareTriple token in
			let oc = open_out_gen [Open_creat; Open_wronly; Open_append] 0o777 ("data.ml") in
			if (!n = 0) then (fprintf oc "open Types;;\n\nlet isCyclic = false;;\n");
			write_data oc res;
			(* fprintf oc "%s" (print_data res); *)
			close_out oc;
			n := !n + 1;
			(* print_string(print_data res) *)
		done		
	with Lexer.Eof ->
		print_endline "Reading input file achieved !"
;;


(* let rec parse oe =
	let lexbuf = Lexing.from_channel (oe) in
	try
		while true do
			let token = Parser.program Lexer.token lexbuf in
			let res = eval_hoareTriple token in
			(* match res with
				| HoareTriple(path,post) -> print_string(print_data(res))
				| InfluenceGraph(ig) -> print_string "Influence Graph\n" *)
			print_string(print_data res)
		done		
	with Lexer.Eof -> 
		print_endline "************\n*** Done ***\n************";
		exit 1
;; *)


let _ =
	let filename = Sys.argv.(1) in
	printf "Input file name : %s\n" filename;
	let oe = open_in filename in
	try 
		parse oe;
	with End_of_file -> ()
;;



